@@ -16,7 +16,7 @@
  * License along with this library; if not, write to the Free
  * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * This file is version 0.9.13 of 17-July-95
+ * This file is version 0.9.14 of 01-November-95
  * Check for the latest version of this package via anonymous ftp at
  *	roxi.rz.fht-mannheim.de:/pub/languages/forth/pfe-VERSION.tar.gz
  * or	sunsite.unc.edu:/pub/languages/forth/pfe-VERSION.tar.gz
@@ -209,8 +209,6 @@ Code (dup_to_r)			/* DUP>R        dito */
   RPUSH (*sp);
 }
 
-Code (random)			/* RANDOM ( n1 --- n2 ) */
-{				/* returns random number within [0,n1) */
 /* some systems (BSD) have a better random number generator than
    standard unix' rand() */
 #if defined HAVE_RANDOM
@@ -219,13 +217,15 @@ Code (random)			/* RANDOM ( n1 --- n2 ) */
 # define _rand_ rand
 #endif
 
+Code (random)			/* RANDOM ( n1 --- n2 ) */
+{				/* returns random number within [0,n1) */
   if (*sp == 0)
     *sp = _rand_ ();
   else
     *sp = ummul (*sp, _rand_ () << 1).hi;
+}
 
 #undef rand
-}
 
 Code (srand)			/* SRAND ( n --- ) */
 {
@@ -1167,6 +1167,16 @@ LISTWORDS (misc) =
   OC ("K8",		EKEY_k8),
   OC ("K9",		EKEY_k9),
   OC ("K10",		EKEY_k0),
+  OC ("S-K1",		EKEY_F1),
+  OC ("S-K2",		EKEY_F2),
+  OC ("S-K3",		EKEY_F3),
+  OC ("S-K4",		EKEY_F4),
+  OC ("S-K5",		EKEY_F5),
+  OC ("S-K6",		EKEY_F6),
+  OC ("S-K7",		EKEY_F7),
+  OC ("S-K8",		EKEY_F8),
+  OC ("S-K9",		EKEY_F9),
+  OC ("S-K10",		EKEY_FA),
   CS ("EXECUTES",	executes),
 
   /* show online help: */
