@@ -16,7 +16,7 @@
  * License along with this library; if not, write to the Free
  * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * This file is version 0.9.13 of 17-July-95
+ * This file is version 0.9.14 of 01-November-95
  * Check for the latest version of this package via anonymous ftp at
  *	roxi.rz.fht-mannheim.de:/pub/languages/forth/pfe-VERSION.tar.gz
  * or	sunsite.unc.edu:/pub/languages/forth/pfe-VERSION.tar.gz
@@ -37,6 +37,7 @@
 #include "dblsub.h"
 #include "term.h"
 #include "lined.h"
+#include "help.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -52,6 +53,10 @@
 #include <unistd.h>		/* access() if available */
 #endif
 
+#if defined HAVE_PWD_H
+#include <pwd.h>		/* to resolve ~user/path */
+#endif
+
 #include "missing.h"
 
 Cell
@@ -361,43 +366,76 @@ try_extensions (char *nm, const char *ext)
   return 0;
 }
 
-#define stpcpy(D,S) (strcpy (D, S) + strlen (S))
+static const char *
+get_home_dir (const char *path, char *home)
+/*
+ * Called when the first character of path is a '~', tries to find
+ * the home directory. Writes the home directory including a trailing
+ * '/' into home and returns the rest of path.
+ */
+{
+  char user[80];
+  path = get_token (path + 1, DIR_DELIMITER, user);
+  if (user[0] == '\0')
+    {
+      char *HOME = getenv ("HOME");
+      sprintf (home, "%s%c", HOME ? HOME : "~", DIR_DELIMITER);
+    }
+  else
+    {
+#if HAVE_PWD_H
+      struct passwd *passwd = getpwnam (user);
+      if (passwd)
+	sprintf (home, "%s%c", passwd->pw_dir, DIR_DELIMITER);
+      else
+#endif
+	sprintf (home, "~%s%c", user, DIR_DELIMITER);
+    }
+  return path;
+}
 
 char *
 expand_filename (const char *nm, const char *paths,
 		 const char *exts, char *fn)
 /*
- * nm file name input, short
- * path search path for files
- * ext  default file extensions
- * fn   full file name, output
+ * nm    file name input, short
+ * paths search path for files
+ * ext   default file extensions
+ * fn    full file name, output
  */
 {
-  char buf[PATH_LENGTH], *p;
-  char *home = getenv ("HOME");
+  char path[PATH_LENGTH];
 
-  if (*nm == '~' && home)
+  if (*nm == DIR_DELIMITER)
     {
-      p = buf;
-      p = stpcpy (p, home);
-      p = stpcpy (p, ++nm);
-      if (try_extensions (buf, exts))
-	return strcpy (fn, buf);
+      strcpy (path, nm);
+      if (try_extensions (path, exts))
+	return strcpy (fn, path);
+    }
+  else if (*nm == '~')
+    {
+      strcat (path, get_home_dir (nm, path));
+      if (try_extensions (path, exts))
+	return strcpy (fn, path);
     }
   else
     {
       while (*paths)
 	{
-	  p = buf;
-	  if (*paths == '~' && home)
+	  char *p;
+	  paths = get_token (paths, PATH_DELIMITER, path);
+	  if (*path == '~')
 	    {
-	      paths++;
-	      p = stpcpy (p, home);
+	      char buf [PATH_LENGTH];
+	      strcpy (buf, path);
+	      strcat (path, get_home_dir (buf, path));
 	    }
-	  paths = get_token (paths, PATH_DELIMITER, p);
-	  strcat (buf, nm);
-	  if (try_extensions (buf, exts))
-	    return strcpy (fn, buf);
+	  p = path + strlen (path);
+	  if (p > path && p[-1] != DIR_DELIMITER)
+	    *p++ = DIR_DELIMITER;
+	  strcpy (p, nm);
+	  if (try_extensions (path, exts))
+	    return strcpy (fn, path);
 	}
     }
   return strcpy (fn, nm);
@@ -623,13 +661,15 @@ to_float (char *p, Cell n, double *r)
     bp,				/* before point, no more sign (had one) */
     ap,				/* after point */
     exn,			/* exponent, maybe sign */
-    ex				/* exponent, no more sign */
+    ex,				/* exponent, no more sign, yet no digit */
+    exd				/* exponent, no more sign, had one digit */
   };
   enum state state = bpn;
   int sign = 1;			/* sign of mantissa */
   long double mant = 0;		/* the mantissa */
   int esign = 1;		/* sign of exponent */
   int exp = 0;			/* the exponent */
+  int bdigs = 0;		/* digits before point */
   int scale = 0;		/* number of digits after point */
 
   while (--n >= 0)
@@ -654,6 +694,7 @@ to_float (char *p, Cell n, double *r)
 		continue;
 	      if (isdigit (c))
 		{
+		  bdigs = 1;
 		  mant = c - '0';
 		  state = bp;
 		  continue;
@@ -666,11 +707,13 @@ to_float (char *p, Cell n, double *r)
 	    case '.':
 	      state = ap;
 	      continue;
+#if 0
 	    case '-':
 	      esign = -1;
 	    case '+':
 	      state = ex;
 	      continue;
+#endif
 	    case 'e':
 	      if (!LOWER_CASE)
 		return 0;
@@ -680,6 +723,7 @@ to_float (char *p, Cell n, double *r)
 	    default:
 	      if (isdigit (c))
 		{
+		  bdigs++;
 		  mant *= 10;
 		  mant += c - '0';
 		  continue;
@@ -711,23 +755,24 @@ to_float (char *p, Cell n, double *r)
 	    }
 	  return 0;
 	case exn:
+	  state = ex;
 	  switch (c)
 	    {
 	    case '-':
 	      esign = -1;
 	    case '+':
-	      state = ex;
 	      continue;
-	    default:
-	      if (isdigit (c))
-		{
-		  exp = c - '0';
-		  state = ex;
-		  continue;
-		}
+	    default: ;
+	    }
+	case ex:
+	  if (isdigit (c))
+	    {
+	      exp = c - '0';
+	      state = exd;
+	      continue;
 	    }
 	  return 0;
-	case ex:
+	case exd:
 	  if (isdigit (c))
 	    {
 	      exp *= 10;
@@ -738,7 +783,7 @@ to_float (char *p, Cell n, double *r)
 	}
     }
   *r = sign * mant * pow10 (scale + esign * exp);
-  return 1;
+  return bdigs - scale > 0;
 }
 
 #endif
@@ -1033,7 +1078,7 @@ question_cr (void)
 void
 file_errorz (const char *fn)
 {
-  tHrow (-256 - errno, fn);
+  tHrow (-1024 - errno, fn);
 }
 
 void
@@ -1241,30 +1286,26 @@ read_line (char *p, uCell *u, File *fid, Cell *ior)
 
   if (!can_read (fid))
     return EPERM;
-  if (feof (fid->f))
-    {
-      *u = 0;
-      *ior = 0;
-      return FALSE;
-    }
   fid->pos = ftell (fid->f);
   for (n = 0; n < *u; n++)
     switch (c = getc (fid->f))
       {
+      default:
+	*p++ = c;
+	continue;
       case EOF:
-	if (!ferror (fid->f))
-	  goto happy;
 	*u = n;
-	*ior = errno;
-	return FALSE;
+	if (ferror (fid->f))
+	  *ior = errno;
+	else
+	  *ior = 0;
+	return FLAG (n > 0);
       case '\r':
 	c = getc (fid->f);
 	if (c != '\n')
 	  ungetc (c, fid->f);
       case '\n':
 	goto happy;
-      default:
-	*p++ = c;
       }
 happy:
   *u = n;
@@ -1293,7 +1334,7 @@ systemf (const char *s,...)	/* issue a system() call after formatting */
 }
 
 /************************************************************************/
-/* source input                                                         */
+/* source input								*/
 /************************************************************************/
 
 /* 1. read from terminal */
@@ -1618,7 +1659,7 @@ word (char del)			/* action of WORD callable from C functions */
 }
 
 /************************************************************************/
-/* inner and outer interpreter                                          */
+/* inner and outer interpreter						*/
 /************************************************************************/
 
 typedef struct			/* jumpbuf is a jmp_buf enhanced by */
@@ -1918,7 +1959,7 @@ unnest_input (Iframe *p)
 }
 
 /************************************************************************/
-/* QUIT, ABORT and exception handling                                   */
+/* QUIT, ABORT and exception handling					*/
 /************************************************************************/
 
 jmp_buf quit_dest;		/* QUIT and ABORT do a THROW which longjmp() */
@@ -1973,7 +2014,7 @@ do_abort (void)
 static void
 show_error (char *fmt,...)
 {
-  char buf[128];
+  char buf[256];
   va_list p;
   int n;
 
@@ -2012,89 +2053,106 @@ throw_msg (int id, char *msg)
 {
   /* *INDENT-OFF* */
   static char *throw_explanation[] =
-  {
-    /*  -1 */ NULL, /* ABORT */
-    /*  -2 */ NULL, /* ABORT" */
-    /*  -3 */ "stack overflow",
-    /*  -4 */ "stack underflow",
-    /*  -5 */ "return stack overflow",
-    /*  -6 */ "return stack underflow",
-    /*  -7 */ "do-loops nested too deeply during execution",
-    /*  -8 */ "dictionary overflow",
-    /*  -9 */ "invalid memory address",
-    /* -10 */ "division by zero",
-    /* -11 */ "result out of range",
-    /* -12 */ "argument type mismatch",
-    /* -13 */ "undefined word",
-    /* -14 */ "interpreting a compile-only word",
-    /* -15 */ "invalid FORGET",
-    /* -16 */ "attempt to use a zero-length string as a name",
-    /* -17 */ "pictured numeric output string overflow",
-    /* -18 */ "parsed string overflow",
-    /* -19 */ "definition name too long",
-    /* -20 */ "write to a read-only location",
-    /* -21 */ "unsupported operation",
-    /* -22 */ "control structure mismatch",
-    /* -23 */ "address alignment exception",
-    /* -24 */ "invalid numeric argument",
-    /* -25 */ "return stack imbalance",
-    /* -26 */ "loop parameters unavailable",
-    /* -27 */ "invalid recursion",
-    /* -28 */ "user interrupt",
-    /* -29 */ "compiler nesting",
-    /* -30 */ "obsolescent feature",
-    /* -31 */ ">BODY used on non-CREATEDd definition",
-    /* -32 */ "invalid name argument",
-    /* -33 */ "block read exception",
-    /* -34 */ "block write exception",
-    /* -35 */ "invalid block number",
-    /* -36 */ "invalid file position",
-    /* -37 */ "file I/O exception",
-    /* -38 */ "non-existent file",
-    /* -39 */ "unexpected end of file",
-    /* -40 */ "invalid BASE for floating-point conversion",
-    /* -41 */ "loss of precision",
-    /* -42 */ "floating-point divide by zero",
-    /* -43 */ "floating-point result out of range",
-    /* -44 */ "floating-point stack overflow",
-    /* -45 */ "floating-point stack underflow",
-    /* -46 */ "floating-point invalid argument",
-    /* -47 */ "compilation word list deleted",
-    /* -48 */ "invalid POSTPONE",
-    /* -49 */ "search-order overflow",
-    /* -50 */ "search-order underflow",
-    /* -51 */ "compilation word list changed",
-    /* -52 */ "control flow stack overflow",
-    /* -53 */ "exception stack overflow",
-    /* -54 */ "floating-point underflow",
-    /* -55 */ "floating-point unidentified fault",
-    /* -56 */ NULL, /* QUIT */
-    /* -57 */ "error in sending or receiving a character",
-    /* -58 */ "[IF], [ELSE] or [THEN] error"
-  }, *pfe_throw_explanation[] =
-  {
-    /* -2048 */ "no or not matching binary image",
-    /* -2049 */ "binary image too big",
-    /* -2050 */ "out of memory",
-    /* -2051 */ "index out of range",
-  };
+    {
+      /*  -1 */ NULL, /* ABORT */
+      /*  -2 */ NULL, /* ABORT" */
+      /*  -3 */ "stack overflow",
+      /*  -4 */ "stack underflow",
+      /*  -5 */ "return stack overflow",
+      /*  -6 */ "return stack underflow",
+      /*  -7 */ "do-loops nested too deeply during execution",
+      /*  -8 */ "dictionary overflow",
+      /*  -9 */ "invalid memory address",
+      /* -10 */ "division by zero",
+      /* -11 */ "result out of range",
+      /* -12 */ "argument type mismatch",
+      /* -13 */ "undefined word",
+      /* -14 */ "interpreting a compile-only word",
+      /* -15 */ "invalid FORGET",
+      /* -16 */ "attempt to use a zero-length string as a name",
+      /* -17 */ "pictured numeric output string overflow",
+      /* -18 */ "parsed string overflow",
+      /* -19 */ "definition name too long",
+      /* -20 */ "write to a read-only location",
+      /* -21 */ "unsupported operation",
+      /* -22 */ "control structure mismatch",
+      /* -23 */ "address alignment exception",
+      /* -24 */ "invalid numeric argument",
+      /* -25 */ "return stack imbalance",
+      /* -26 */ "loop parameters unavailable",
+      /* -27 */ "invalid recursion",
+      /* -28 */ "user interrupt",
+      /* -29 */ "compiler nesting",
+      /* -30 */ "obsolescent feature",
+      /* -31 */ ">BODY used on non-CREATEDd definition",
+      /* -32 */ "invalid name argument",
+      /* -33 */ "block read exception",
+      /* -34 */ "block write exception",
+      /* -35 */ "invalid block number",
+      /* -36 */ "invalid file position",
+      /* -37 */ "file I/O exception",
+      /* -38 */ "non-existent file",
+      /* -39 */ "unexpected end of file",
+      /* -40 */ "invalid BASE for floating-point conversion",
+      /* -41 */ "loss of precision",
+      /* -42 */ "floating-point divide by zero",
+      /* -43 */ "floating-point result out of range",
+      /* -44 */ "floating-point stack overflow",
+      /* -45 */ "floating-point stack underflow",
+      /* -46 */ "floating-point invalid argument",
+      /* -47 */ "compilation word list deleted",
+      /* -48 */ "invalid POSTPONE",
+      /* -49 */ "search-order overflow",
+      /* -50 */ "search-order underflow",
+      /* -51 */ "compilation word list changed",
+      /* -52 */ "control flow stack overflow",
+      /* -53 */ "exception stack overflow",
+      /* -54 */ "floating-point underflow",
+      /* -55 */ "floating-point unidentified fault",
+      /* -56 */ NULL, /* QUIT */
+      /* -57 */ "error in sending or receiving a character",
+      /* -58 */ "[IF], [ELSE] or [THEN] error"
+    },
+  *pfe_throw_explanation[] =
+    {
+      /* -2048 */ "no or not matching binary image",
+      /* -2049 */ "binary image too big",
+      /* -2050 */ "out of memory",
+      /* -2051 */ "index out of range",
+    };
   /* *INDENT-ON* */
 
   if (-1 - DIM (throw_explanation) < id && id <= -1)
-    strcpy (msg, throw_explanation[-1 - id]);
-  else if (-2048 < id && id <= -256)
     {
+      /* ANS-Forth throw codes, messages are in throw_explanation[] */
+      strcpy (msg, throw_explanation[-1 - id]);
+    }
+  else if (-1024 < id && id <= -256)
+    {
+      /* Signals, see signal.c, those not handled and not fatal lead
+         to THROW */
+      char sigmsg [0x80];
+      strcpy (sigmsg, msg);
+      sprintf (msg, "Received signal %d, %s", -256 - id, sigmsg);
+    }
+  else if (-2048 < id && id <= -1024)
+    {
+      /* File errors */
       char fn[PATH_LENGTH];
-
       strcpy (fn, msg);
-      sprintf (msg, ": File %s: %s", fn, strerror (-256 - id));
+      sprintf (msg, ": File %s: %s", fn, strerror (-1024 - id));
     }
   else if (-2048 - DIM (pfe_throw_explanation) < id && id <= -2048)
-    strcpy (msg, pfe_throw_explanation[-2048 - id]);
+    {
+      strcpy (msg, pfe_throw_explanation[-2048 - id]);
+    }
   else
-    sprintf (msg, "%d THROW unassigned", id);
+    {
+      sprintf (msg, "%d THROW unassigned", id);
+    }
 }
 
+
 int
 cAtch (Xt xt)
 {
@@ -2122,7 +2180,7 @@ tHrow (int id,...)
 {
   Except *x = sys.cAtch;
   va_list p;
-  char msg[80];
+  char msg[256];
 
   if (x && x->magic == EXCEPTION_MAGIC)
     {
@@ -2190,7 +2248,7 @@ question_file_open (File *fid)
 }
 
 /************************************************************************/
-/* Initialize dictionary, and system variables, include files           */
+/* Initialize dictionary, and system variables, include files		*/
 /************************************************************************/
 
 static void
@@ -2226,8 +2284,16 @@ init_dictionary (Dict *dict, uCell size)
   quit_initializations ();
 
   /* Include .pferc if it exists: */
-  if (option.pferc_file && access (option.pferc_file, R_OK) == 0)
-    included (option.pferc_file, strlen (option.pferc_file));
+  if (option.pferc_file)
+    {
+      char pferc [PATH_LENGTH];
+      if (*option.pferc_file == '~')
+	strcat (pferc, get_home_dir (option.pferc_file, pferc));
+      else
+	strcpy (pferc, option.pferc_file);
+      if (access (pferc, R_OK) == 0)
+	included (pferc, strlen (pferc));
+    }
 
   /* Include file from command line: */
   if (option.include_file)
