@@ -16,7 +16,7 @@
  * License along with this library; if not, write to the Free
  * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * This file is version 0.9.13 of 17-July-95
+ * This file is version 0.9.14 of 01-November-95
  * Check for the latest version of this package via anonymous ftp at
  *	roxi.rz.fht-mannheim.de:/pub/languages/forth/pfe-VERSION.tar.gz
  * or	sunsite.unc.edu:/pub/languages/forth/pfe-VERSION.tar.gz
@@ -154,6 +154,9 @@ no_waitchar (int fd)		/* "don't wait for character" */
 #ifndef VQUIT
 #define VQUIT 1
 #endif
+#ifndef VERASE
+#define VERASE 2
+#endif
 #ifndef VMIN
 #define VMIN 4
 #endif
@@ -171,6 +174,7 @@ static struct termios tty_system;
 #else
 # define tty_ospeed	(tty_system.c_ospeed)
 #endif
+#define tty_erasechar	(tty_system.c_cc[VERASE])
 
 int
 tty_interrupt_key (char ch)
@@ -258,6 +262,7 @@ tty_system;
 			ioctl (STDIN_FILENO, TIOCSETC, &tty_system.tc), \
 			ioctl (STDIN_FILENO, TIOCLSET, &tty_system.lc)
 #define tty_ospeed	(tty_system.sg.sg_ospeed)
+#define tty_erasechar	(tty_system.sg.sg_erase)
 
 int
 tty_interrupt_key (char ch)
@@ -285,7 +290,7 @@ tty_interrupt_key (char ch)
 
 #define	C_FLAGS_OFF	(LCASE | ECHO | RAW)
 #define C_FLAGS_ON	(CBREAK)
-#define L_FLAGS_ON	(LTOSTOP | LPASS8OUT)
+#define L_FLAGS_ON	(LTOSTOP | LPASS8)
 
 static void
 tty_interactive (void)
@@ -386,14 +391,16 @@ enum
 /*
  * Strings used to query the termcap database.
  *
- * (In case terminfo is * used, these are only needed to satisfy the
- * show_*_strings * functions.)
+ * (In case terminfo is used, these are only needed to satisfy the
+ * show_*_strings functions.)
  */
 
 static char tckeycode[][3] =
 {
   "k1", "k2", "k3", "k4", "k5",	/* keys in same order as enum keycode */
   "k6", "k7", "k8", "k9", "k0",	/* from term.h */
+  "F1", "F2", "F3", "F4", "F5",
+  "F6", "F7", "F8", "F9", "FA",
   "kl", "kr", "ku", "kd",	/* result is just what has to be exported */
   "kh", "kH", "kN", "kP",	/* via variable rawkey_string */
   "kb", "kD", "kM", "kI",
@@ -466,6 +473,17 @@ char *rawkey_string[] =		/* Strings sent by function keys */
   "\033[20~",			/* k9 */
   "\033[21~",			/* k0 */
 
+  "\033[23~",			/* F1 - function keys S-F1 thru S-F10 */
+  "\033[24~",			/* F2 - stem from Linux console */
+  "\033[25~",			/* F3 - or whoknowswhereelse */
+  "\033[26~",			/* F4 */
+  "\033[28~",			/* F5 */
+  "\033[29~",			/* F6 */
+  "\033[31~",			/* F7 */
+  "\033[32~",			/* F8 */
+  "\033[33~",			/* F9 */
+  "\033[34~",			/* FA */
+
   "\033OD",			/* kl - arrow left */
   "\033OC",			/* kr - arrow right */
   "\033OA",			/* ku - arrow up */
@@ -548,6 +566,7 @@ query_database (void)
   char *ttype = getenv ("TERM");
   char tcent[2048];
   static char tcbuf[2048];
+  static char erase[] = "\b";
   char *tctop = tcbuf;
   int i;
 
@@ -570,7 +589,16 @@ query_database (void)
   /* Read all termcap strings we need, */
   for (i = 0; i < NO_OF_KEYS; i++)
     rawkey_string[i] = tgetstr (tckeycode[i], &tctop);
-  /* rawkey_string [EKEY_enter - EKEY_k1] = "\r"; */
+
+  /* another chance for F10: */
+  if (rawkey_string [EKEY_k0 - EKEY_k1] == NULL)
+    rawkey_string [EKEY_k0 - EKEY_k1] = tgetstr ("k;", &tctop);
+  /* better believe stty which is the erase character: */
+  if (tty_erasechar)
+    {
+      erase[0] = tty_erasechar;
+      rawkey_string [EKEY_kb - EKEY_k1] = erase;
+    }
   for (i = 0; i < DIM (tcctlcode); i++)
     control_string[i] = tgetstr (tcctlcode[i], &tctop);
   if (control_string [cursor_left] == NULL)
@@ -613,6 +641,11 @@ t_puts (int tcidx, int n)	/* issue termcap string to terminal */
  * 2.3 Use terminfo calls to retrieve information about the terminal
  */
 
+#if SunOS			/* I tried Solaris 2.4 fr x86 */
+#undef TRUE			/* These confuse Sun's cc in curses.h */
+#undef FALSE
+#include <sys/termio.h>		/* Solaris' curses.h seems to need this */
+#endif
 #include <curses.h>
 #include <term.h>
 
@@ -623,6 +656,7 @@ static int
 query_database (void)
 {
   int errret;
+  static char erase[] = "\b";
 
   setupterm (NULL, STDOUT_FILENO, &errret);
   if (errret != 1)
@@ -663,6 +697,13 @@ query_database (void)
   KINIT (kC, key_clear);	/* clear screen key */
 
 #undef KINIT
+  /* better believe stty which is the erase character: */
+  if (tty_erasechar)
+    {
+      erase[0] = tty_erasechar;
+      rawkey_string [EKEY_kb - EKEY_k1] = erase;
+    }
+
   if (cursor_left == NULL)
     cursor_left = "\b";
   return 1;
